CPU_EVENT(frameRead)
CPU_EVENT(bitstreamWrite)
CPU_EVENT(frameThread)
CPU_EVENT(encodeCTU)
CPU_EVENT(filterCTURow)
CPU_EVENT(slicetypeDecideEV)
CPU_EVENT(prelookahead)
CPU_EVENT(estCostSingle)
CPU_EVENT(estCostCoop)
CPU_EVENT(pmode)
CPU_EVENT(pme)
