#ifndef PIXEL_PRIM_NEON_H__
#define PIXEL_PRIM_NEON_H__

#include "common.h"
#include "slicetype.h"      // LOWRES_COST_MASK
#include "primitives.h"
#include "x265.h"



namespace X265_NS
{



void setupPixelPrimitives_neon(EncoderPrimitives &p);


}


#endif

